from __future__ import print_function

import sys
import os

import _setup_test_env  # noqa
from nose.tools import assert_equal
import returnn.sprint.interface as SprintAPI
from returnn.tf.engine import Engine
from tempfile import mkdtemp
from returnn.config import Config
import shutil
import numpy


def install_sigint_handler():
  import signal

  def signal_handler(signal, frame):
    print("\nSIGINT at:")
    better_exchook.print_tb(tb=frame, file=sys.stdout)
    print("")

    # It's likely that SIGINT was caused by Util.interrupt_main().
    # We might have a stacktrace from there.
    if getattr(sys, "exited_frame", None) is not None:
      print("interrupt_main via:")
      better_exchook.print_tb(tb=sys.exited_frame, file=sys.stdout)
      print("")
      sys.exited_frame = None
      # Normal exception instead so that Nose will catch it.
      raise Exception("Got SIGINT!")
    else:
      print("\nno sys.exited_frame\n")
      # Normal SIGINT. Normal Nose exit.
      if old_action:
        old_action()
      else:
        raise KeyboardInterrupt

  old_action = signal.signal(signal.SIGINT, signal_handler)


install_sigint_handler()


def test_forward():
  tmpdir = mkdtemp("returnn-test-sprint")
  olddir = os.getcwd()
  os.chdir(tmpdir)

  from returnn.datasets.generating import DummyDataset
  seq_len = 5
  n_data_dim = 2
  n_classes_dim = 3
  train_data = DummyDataset(input_dim=n_data_dim, output_dim=n_classes_dim, num_seqs=4, seq_len=seq_len)
  train_data.init_seq_order(epoch=1)
  cv_data = DummyDataset(input_dim=n_data_dim, output_dim=n_classes_dim, num_seqs=2, seq_len=seq_len)
  cv_data.init_seq_order(epoch=1)

  config = "\n".join([
    "#!rnn.py",
    "model = '%s/model'" % tmpdir,
    "num_outputs = %i" % n_classes_dim,
    "num_inputs = %i" % n_data_dim,
    'network = {"output": {"class": "softmax", "loss": "ce", "from": "data:data"}}',
    "num_epochs = 2",
  ])

  open("config_write", "w").write(config)
  open("config", "w").write(config + "\nload= '%s/model'" % tmpdir)

  config = Config()
  config.load_file("config_write")
  engine = Engine(config=config)
  engine.init_train_from_config(config=config, train_data=train_data, dev_data=cv_data, eval_data=None)
  engine.epoch = 1
  engine.save_model(engine.get_epoch_model_filename())
  Engine._epoch_model = None

  # Reset engine
  from returnn.util.basic import BackendEngine
  BackendEngine.selected_engine = None

  inputDim = 2
  outputDim = 3
  SprintAPI.init(inputDim=inputDim, outputDim=outputDim,
                 config="action:forward,configfile:config,epoch:1",
                 targetMode="forward-only")
  assert isinstance(SprintAPI.engine, Engine)
  print("used data keys via net:", SprintAPI.engine.network.get_used_data_keys())

  features = numpy.array([[0.1, 0.2], [0.2, 0.3], [0.3, 0.4], [0.4, 0.5]])
  seq_len = features.shape[0]
  posteriors = SprintAPI._forward("segment1", features.T).T
  assert_equal(posteriors.shape, (seq_len, outputDim))

  SprintAPI.exit()

  os.chdir(olddir)
  shutil.rmtree(tmpdir)



"""
Main engine for PyTorch
"""

from typing import Optional, Callable

import torch
import numpy
import os
from torch.utils.data import DataLoader
from random import random

from returnn.log import log
from returnn.engine.base import EngineBase
from returnn.learning_rate_control import load_learning_rate_control_from_config, LearningRateControl
from returnn.datasets.basic import init_dataset
from returnn.util import basic as util
from . import data_pipeline


class Engine(EngineBase):
  """
  PyTorch engine
  """

  def __init__(self, config):
    """
    :param returnn.config.Config config:
    """
    super(Engine, self).__init__()
    self.config = config
    self.model_filename = self.config.value('model', None)
    self.train_dataset = None
    self.eval_datasets = {}

    self._start_epoch = None  # type: Optional[int]
    self._final_epoch = None  # type: Optional[int]
    self._model = None  # type: Optional[torch.nn.Module]
    self._train_step_func = None  # type: Optional[Callable]
    self._learning_rate = 0.0
    self._learning_rate_control = None  # type: Optional[LearningRateControl]
    self._save_model_epoch_interval = 1
    self._optimizer = None  # type: Optional[torch.optim.Optimizer]

    self._device = "cuda" if torch.cuda.is_available() else "cpu"

  def init_train_from_config(self, config=None, train_data=None, dev_data=None, eval_data=None):
    """
    :param returnn.config.Config|None config:
    :param returnn.datasets.basic.Dataset|None train_data:
    :param returnn.datasets.basic.Dataset|None dev_data:
    :param returnn.datasets.basic.Dataset|None eval_data:
    """
    assert config is self.config
    self.train_dataset = train_data
    self.eval_datasets.clear()
    if dev_data:
      self.eval_datasets["dev"] = dev_data
    if eval_data:
      self.eval_datasets["eval"] = eval_data
    if config.has("eval_datasets"):
      for dataset_name, dataset_opts in config.typed_value("eval_datasets", {}).items():
        self.eval_datasets[dataset_name] = init_dataset(dataset_opts, default_kwargs={"name": dataset_name})

    self._start_epoch, _ = self.get_train_start_epoch_batch(self.config)
    self._final_epoch = self.config_get_final_epoch(self.config)

    self._load_model(epoch=self._start_epoch)
    self._learning_rate_control = load_learning_rate_control_from_config(config)
    self._save_model_epoch_interval = config.int('save_interval', 1)

    if self._start_epoch > 1:
      self._load_optimizer(epoch=self._start_epoch)
    else:
      # TODO: configure according to config
      self._optimizer = torch.optim.Adam(self._model.parameters(), lr=self._learning_rate)

    self._train_step_func = self.config.typed_value("train_step")
    assert self._train_step_func, "train_step not defined"

  def train(self):
    """
    Main training loop.
    """

    print("Starting training at epoch {}.".format(self._start_epoch), file=log.v3)
    assert self._model, "Model not initialized, call init_train_from_config()."

    self.epoch = self._start_epoch
    while self.epoch <= self._final_epoch:
      self.init_train_epoch()
      self.train_epoch()

      self.epoch += 1

    print("Finished training at epoch {}.".format(self.epoch), file=log.v3)

  def init_train_epoch(self):
    """
    init train (sub)epoch. LR etc
    """
    self._learning_rate = self._learning_rate_control.get_learning_rate_for_epoch(self.epoch)

    # Update learning rate
    for param_group in self._optimizer.param_groups:
      param_group["lr"] = self._learning_rate

  def train_epoch(self):
    """
    train one (sub)epoch
    """
    print("start", self.get_epoch_str(), "with learning rate", self._learning_rate, "...", file=log.v4)

    data_loader = self._create_data_loader(self.train_dataset)

    self._model.train()

    step_idx = 0
    for data in data_loader:
      loss = self._run_step(data)

      self._optimizer.zero_grad()
      loss.backward()
      self._optimizer.step()

      print("step %i, loss: %f" % (step_idx, loss.detach().cpu().numpy()), file=log.v4)

      step_idx += 1

    print("Trained %i steps" % step_idx)

    if self.epoch % self._save_model_epoch_interval == 0 or self.epoch == self._final_epoch:
      self._save_model()
      self._save_optimizer()

    self.eval_model()

  def eval_model(self):
    """
    Runs model on all eval datasets and calculates the loss.
    """
    self._model.eval()

    for dataset_name, dataset in self.eval_datasets.items():
      print("Evaluating dataset '{}'...".format(dataset_name), file=log.v3)

      data_loader = self._create_data_loader(dataset)

      accumulated_loss = 0.0
      step_idx = 0

      with torch.no_grad():
        for data in data_loader:

          loss = self._run_step(data).detach().cpu().numpy()
          print("step %i, loss: %f" % (step_idx, loss), file=log.v4)

          accumulated_loss += loss
          step_idx += 1

      total_loss = accumulated_loss / step_idx if step_idx else 0.0

      self._learning_rate_control.set_epoch_error(self.epoch, {"%s_score" % dataset_name: float(total_loss)})
      self._learning_rate_control.save()

      print("Total loss for '{}': {:.6}".format(dataset_name, total_loss), file=log.v3)

  def _create_data_loader(self, dataset):
    """
    :param returnn.datasets.basic.Dataset dataset:
    :return: PyTorch data loader created from given RETURNN dataset
    :rtype: DataLoader
    """
    wrapped_dataset = data_pipeline.DatasetWrapper(dataset, epoch=self.epoch)

    batch_max_seqs = self.config.int('max_seqs', 1)  # TODO wrong default, actually -1, no limit (limit via batch_size)
    data_loader = DataLoader(
      wrapped_dataset,
      batch_size=batch_max_seqs,
      collate_fn=data_pipeline.collate_batch,
    )

    return data_loader

  def _run_step(self, data):
    """
    :param dict[str, numpy.ndarray] data: model inputs for the step
    :return: model loss calculated for the step
    :rtype: torch.Tensor
    """
    assert isinstance(data, dict) and data
    data = {
      k:
      v.cpu() if k.endswith(":seq_len")
      else v.to(self._device)
      for (k, v) in data.items()
    }  # Sequence lengths have to be on CPU for the later call to rnn.pack_padded_sequence

    train_ctx = TrainCtx()
    sentinel_kw = {"__fwd_compatible_random_arg_%i" % int(random() * 100): None}
    self._train_step_func(model=self._model, data=data, train_ctx=train_ctx, **sentinel_kw)
    loss = train_ctx.total_loss()

    return loss

  def _load_model(self, epoch):
    """
    Sets self._model to a torch.nn.Module.

    :param int epoch:
    """
    get_model_func = self.config.typed_value("get_model")
    assert get_model_func, "get_model not defined"
    sentinel_kw = {"__fwd_compatible_random_arg_%i" % int(random() * 100): None}
    self._model = get_model_func(**sentinel_kw)
    assert isinstance(self._model, torch.nn.Module)

    if epoch > 1:
      filename = self.get_epoch_model_filename(epoch=epoch - 1) + util.get_model_filename_postfix()
      print("Load model %s" % (filename,), file=log.v4)
      model_state = torch.load(filename)
      self._model.load_state_dict(model_state)

    self._model.to(self._device)

  def _save_model(self):
    """
    Saves the state of self._model to file.
    """
    filename = self.get_epoch_model_filename() + util.get_model_filename_postfix()
    directory = os.path.dirname(filename)
    if not os.path.exists(directory):
      os.makedirs(directory, exist_ok=True)

    print("Save model under %s" % (filename,), file=log.v4)
    torch.save(self._model.state_dict(), filename)

  def _load_optimizer(self, epoch):
    """
    Loads a torch.optim.Optimizer from disk and stores it in self._optimizer.
    """
    filename = self.get_epoch_model_filename(epoch=epoch - 1) + ".opt" + util.get_model_filename_postfix()
    print("Load optimizer %s" % filename, file=log.v4)
    optimizer_state = torch.load(filename)
    self._optimizer.load_state_dict(optimizer_state)

  def _save_optimizer(self):
    """
    Saves the state of self._optimizer to file.
    """
    filename = self.get_epoch_model_filename() + ".opt" + util.get_model_filename_postfix()
    directory = os.path.dirname(filename)
    if not os.path.exists(directory):
      os.makedirs(directory, exist_ok=True)

    print("Save optimizer under %s" % filename, file=log.v4)
    torch.save(self._optimizer.state_dict(), filename)


class TrainCtx:
  """
  train ctx
  """

  def __init__(self):
    self.loss = None

  def mark_as_loss(self, loss):
    """
    :param torch.Tensor loss:
    """
    assert self.loss is None, "don't call multiple times"
    self.loss = loss

  def total_loss(self):
    """
    :rtype: torch.Tensor
    """
    assert self.loss is not None, "call train_ctx.mark_as_loss"
    return self.loss




"""
Provides the learning rate scheduling logic.
The base class is :class:`LearningRateControl`.
"""

from __future__ import print_function

import os
import typing
from returnn.util.basic import better_repr, simple_obj_repr, ObjAsDict, unicode
from returnn.log import log
import numpy


class LearningRateControl(object):
  """
  Base class for learning rate control / scheduling.
  """

  need_error_info = True

  class EpochData:
    """
    Encapsulates all relevant information for one epoch,
    needed to perform learning rate scheduling,
    such as the individual scores (cv or train; cross-entropy or frame-error or whatever).
    """

    # Need to keep the non-PEP8 name for compatibility, because we store the repr of the object.
    # noinspection PyPep8Naming
    def __init__(self, learningRate, error=None):
      """
      :type learningRate: float
      :type error: dict[str,float] | None
      """
      self.learning_rate = learningRate
      if isinstance(error, float):  # Old format.
        error = {"old_format_score": error}
      if error is None:
        error = {}
      self.error = error

    def __repr__(self):
      # This is being used for serialization, and we want some forward/backward compatibility,
      # so we should try to keep this consistent.
      return "EpochData(learningRate=%s, error=%s)" % (
        better_repr(self.learning_rate), better_repr(self.error))

  @classmethod
  def load_initial_kwargs_from_config(cls, config):
    """
    :type config: returnn.config.Config
    :rtype: dict[str]
    """
    return {
      "default_learning_rate": config.float('learning_rate', 1.0),
      "min_learning_rate": config.float('min_learning_rate', 0.0),
      "default_learning_rates": config.typed_value('learning_rates') or config.float_list('learning_rates'),
      "error_measure_key": (
        config.typed_value('learning_rate_control_error_measure')
        or config.value('learning_rate_control_error_measure', None)),
      "relative_error_also_relative_to_learning_rate": (
        config.bool('learning_rate_control_relative_error_relative_lr', False)),
      "min_num_epochs_per_new_learning_rate": config.int("learning_rate_control_min_num_epochs_per_new_lr", 0),
      "relative_error_div_by_old": config.bool('newbob_relative_error_div_by_old', False),
      "learning_rate_decay": config.typed_value(
        'learning_rate_decay', config.opt_typed_value('newbob_learning_rate_decay', 0.5)),
      "learning_rate_growth": config.typed_value(
        'learning_rate_growth', config.opt_typed_value('newbob_learning_rate_growth', 1.0)),
      "filename": config.value('learning_rate_file', None),
    }

  @classmethod
  def load_initial_from_config(cls, config):
    """
    :type config: returnn.config.Config
    :rtype: LearningRateControl
    """
    kwargs = cls.load_initial_kwargs_from_config(config)
    return cls(**kwargs)

  def __init__(self, default_learning_rate, min_learning_rate=0.0, default_learning_rates=None,
               error_measure_key=None,
               relative_error_also_relative_to_learning_rate=False,
               min_num_epochs_per_new_learning_rate=0,
               relative_error_div_by_old=False,
               learning_rate_decay=1.0,
               learning_rate_growth=1.0,
               filename=None):
    """
    :param float default_learning_rate: default learning rate. usually for epoch 1
    :param list[float] | dict[int,float] default_learning_rates: learning rates
    :param str|list[str]|None error_measure_key: for get_epoch_error_value() the key for EpochData.error which is a dict
    :param int min_num_epochs_per_new_learning_rate: if the lr was recently updated, use it for at least N epochs
    :param bool relative_error_div_by_old: if True, compute relative error as (new - old) / old.
    :param float|(float)->float learning_rate_decay:
    :param float|(float)->float learning_rate_growth:
    :param str filename: load from and save to file
    """
    self.epoch_data = {}  # type: typing.Dict[int,LearningRateControl.EpochData]
    self.filename = filename
    if filename:
      if os.path.exists(filename):
        print("Learning-rate-control: loading file %s" % filename, file=log.v4)
        # Load now, such that default_learning_rates is correctly handled.
        self.load()
      else:
        print("Learning-rate-control: file %s does not exist yet" % filename, file=log.v4)
    else:
      print("Learning-rate-control: no file specified, not saving history (no proper restart possible)", file=log.v4)
    self.default_learning_rate = default_learning_rate
    self.min_learning_rate = min_learning_rate
    if default_learning_rates:
      if isinstance(default_learning_rates, list):
        default_learning_rates = {i + 1: v for (i, v) in enumerate(default_learning_rates)}
      if isinstance(default_learning_rates, (str, unicode)):
        default_learning_rates = eval(default_learning_rates)
      assert isinstance(default_learning_rates, dict)
      for epoch, v in default_learning_rates.items():
        self.set_default_learning_rate_for_epoch(epoch, v)
    self.default_learning_rates = default_learning_rates
    self.error_measure_key = error_measure_key
    self.relative_error_also_relative_to_learning_rate = relative_error_also_relative_to_learning_rate
    self.min_num_epochs_per_new_learning_rate = min_num_epochs_per_new_learning_rate
    self.relative_error_div_by_old = relative_error_div_by_old
    self.learning_rate_decay = learning_rate_decay
    self.learning_rate_growth = learning_rate_growth

  __repr__ = simple_obj_repr

  def __str__(self):
    epochs = sorted(self.epoch_data.keys())
    if len(epochs) > 6:
      epoch_str = ", ".join(
        ["%i: %s" % (epoch, self.epoch_data[epoch]) for epoch in epochs[:3]] +
        ["..."] +
        ["%i: %s" % (epoch, self.epoch_data[epoch]) for epoch in epochs[-3:]])
    else:
      epoch_str = ", ".join(["%i: %s" % (epoch, self.epoch_data[epoch]) for epoch in epochs])
    return "%r, epoch data: %s, error key: %s" % (self, epoch_str, self.get_error_key(epoch=1))

  @staticmethod
  def _calc_learning_rate_update(learning_rate, update):
    """
    :param float learning_rate:
    :param None|float|(float)->float update: factor, or generic func
    :return: lr with update applied (e.g. decay factor)
    :rtype: float
    """
    if update is None:
      return learning_rate
    if isinstance(update, float):
      return learning_rate * update
    assert callable(update)
    learning_rate = update(learning_rate)
    assert isinstance(learning_rate, float)
    return learning_rate

  def _calc_learning_rate_decay(self, learning_rate):
    """
    :param float learning_rate:
    :return: lr with decay applied
    :rtype: float
    """
    return self._calc_learning_rate_update(learning_rate, update=self.learning_rate_decay)

  def _calc_learning_rate_growth(self, learning_rate):
    """
    :param float learning_rate:
    :return: lr with growth applied
    :rtype: float
    """
    return self._calc_learning_rate_update(learning_rate, update=self.learning_rate_growth)

  def calc_learning_rate_decay_or_grow(self, learning_rate, decay, grow=None):
    """
    :param float learning_rate:
    :param bool decay:
    :param bool|None grow: default is not decay
    :return: lr with decay or growth applied
    :rtype: float
    """
    assert isinstance(decay, bool)
    if grow is None:
      grow = not decay
    assert isinstance(grow, bool)
    assert not (grow and decay)  # not sure if this makes sense...
    if decay:
      learning_rate = self._calc_learning_rate_decay(learning_rate)
      if learning_rate < self.min_learning_rate:
        learning_rate = self.min_learning_rate
    if grow:
      learning_rate = self._calc_learning_rate_growth(learning_rate)
    return learning_rate

  def calc_learning_rate_for_epoch(self, epoch):
    """
    :type epoch: int
    :returns learning rate
    :rtype: float
    """
    raise NotImplementedError

  def calc_new_learning_rate_for_epoch(self, epoch):
    """
    :param int epoch:
    :return: new learning rate for this epoch
    :rtype: float
    """
    if self.min_num_epochs_per_new_learning_rate > 1:
      last_lrs = [self.epoch_data[e].learning_rate
                  for e in self._last_epochs_for_epoch(epoch, num_epochs=self.min_num_epochs_per_new_learning_rate)]
      if len(set(last_lrs)) >= 2 or 0 < len(last_lrs) < self.min_num_epochs_per_new_learning_rate:
        return last_lrs[-1]
    learning_rate = self.calc_learning_rate_for_epoch(epoch)
    return learning_rate

  def _last_epochs_for_epoch(self, epoch, num_epochs):
    """
    :param int epoch:
    :param int num_epochs:
    :return: last N epochs where we have some epoch data
    :rtype: list[int]
    """
    last_epochs = sorted([e for e in self.epoch_data.keys() if e < epoch])
    if not last_epochs:
      return []
    last_epochs = last_epochs[-num_epochs:]
    return last_epochs

  def get_learning_rate_for_epoch(self, epoch):
    """
    :type epoch: int
    :rtype: float
    """
    assert epoch >= 1
    if epoch in self.epoch_data:
      return self.epoch_data[epoch].learning_rate
    learning_rate = self.calc_new_learning_rate_for_epoch(epoch)
    self.set_default_learning_rate_for_epoch(epoch, learning_rate)
    return learning_rate

  def set_default_learning_rate_for_epoch(self, epoch, learning_rate):
    """
    :type epoch: int
    :type learning_rate: float
    """
    if epoch in self.epoch_data:
      if not self.epoch_data[epoch].learning_rate:
        self.epoch_data[epoch].learning_rate = learning_rate
    else:
      self.epoch_data[epoch] = self.EpochData(learning_rate)

  def get_last_epoch(self, epoch):
    """
    :param int epoch:
    :return: last epoch before ``epoch`` where we have some epoch data
    :rtype: int
    """
    epochs = sorted([e for e in self.epoch_data.keys() if e < epoch])
    if not epochs:
      return None
    return epochs[-1]

  def get_most_recent_learning_rate(self, epoch, exclude_current=True):
    """
    :param int epoch:
    :param bool exclude_current:
    :return: most learning rate before or including ``epoch``
    :rtype: float
    """
    for e, data in reversed(sorted(self.epoch_data.items())):
      assert isinstance(data, LearningRateControl.EpochData)
      if e > epoch:
        continue
      if exclude_current and e == epoch:
        continue
      if data.learning_rate is None:
        continue
      return data.learning_rate
    return self.default_learning_rate

  def calc_relative_error(self, old_epoch, new_epoch):
    """
    :param int old_epoch:
    :param int new_epoch:
    :return: relative error between old epoch and new epoch
    :rtype: float
    """
    old_key, old_error = self.get_epoch_error_key_value(old_epoch)
    new_key, new_error = self.get_epoch_error_key_value(new_epoch)
    if old_error is None or new_error is None:
      return None
    if old_key != new_key:
      return None
    if self.relative_error_div_by_old:
      relative_error = (new_error - old_error) / abs(old_error)
    else:
      relative_error = (new_error - old_error) / abs(new_error)
    if self.relative_error_also_relative_to_learning_rate:
      learning_rate = self.get_most_recent_learning_rate(new_epoch, exclude_current=False)
      if learning_rate > 0:
        # If the learning rate is lower than the initial learning rate,
        # the relative error is also expected to be lower, so correct for that here.
        relative_error /= learning_rate / self.default_learning_rate
    return relative_error

  def set_epoch_error(self, epoch, error):
    """
    :type epoch: int
    :type error: dict[str,float|dict[str,float]]
    """
    if epoch not in self.epoch_data:
      print("Learning rate not set for epoch %i. Assuming default." % epoch, file=log.v4)
      self.get_learning_rate_for_epoch(epoch)  # This will set it.
    assert isinstance(error, dict)
    error = error.copy()
    for k, v in list(error.items()):
      if isinstance(v, dict):  # like error = {"dev_score": {"cost:output1": .., "cost:output2": ...}, ...}
        del error[k]
        if len(v) == 1:
          error[k] = list(v.values())[0]
          continue
        for k1, v1 in v.items():
          if ":" in k1:
            k1 = k1[k1.index(":") + 1:]
          error[k + "_" + k1] = v1
    for v in error.values():
      assert isinstance(v, float)
    self.epoch_data[epoch].error.update(error)
    if epoch == 1:
      print("Learning-rate-control: error key %r from %r" % (self.get_error_key(epoch), error), file=log.v4)

  def get_error_key(self, epoch):
    """
    :param int epoch:
    :return: key which we should look in scores/errors, for this epoch
    :rtype: str
    """
    if epoch not in self.epoch_data:
      if isinstance(self.error_measure_key, list):
        return self.error_measure_key[0]
      assert isinstance(self.error_measure_key, (str, type(None)))
      return self.error_measure_key
    epoch_data = self.epoch_data[epoch]
    if not epoch_data.error:
      return None
    if len(epoch_data.error) == 1 and "old_format_score" in epoch_data.error:
      return "old_format_score"
    keys = []
    if isinstance(self.error_measure_key, list):
      for key in self.error_measure_key:
        keys += [key, key + "_output"]  # for multiple outputs, try default output
    elif isinstance(self.error_measure_key, str):
      keys += [self.error_measure_key, self.error_measure_key + "_output"]
    else:
      assert self.error_measure_key is None
    keys += ["dev_score", "dev_score_output"]
    for key in keys:
      if key in epoch_data.error:
        return key
    for key in sorted(epoch_data.error.keys()):
      if key == "dev_score_output/output" or key.startswith("dev_score_output/output_"):
        return key
    for key in sorted(epoch_data.error.keys()):
      if key.startswith("dev_score_output/"):
        return key
    for key in sorted(epoch_data.error.keys()):
      if key.startswith("dev_"):
        return key
    for key in ["train_score", "train_score_output"]:
      if key in epoch_data.error:
        return key
    return min(epoch_data.error.keys())

  def get_epoch_error_dict(self, epoch):
    """
    :param int epoch:
    :rtype: dict[str,float]
    """
    if epoch not in self.epoch_data:
      return {}
    return self.epoch_data[epoch].error

  def get_epoch_error_value(self, epoch):
    """
    :param int epoch:
    :return: error/score for the specific epoch, given the error-key, see :func:`get_error_key`
    :rtype: float
    """
    error = self.get_epoch_error_dict(epoch)
    if not error:
      return None
    key = self.get_error_key(epoch)
    assert key
    assert key in error, (
      "%r not in %r. fix %r in config. set it to %r or so." % (
        key, error, 'learning_rate_control_error_measure', 'dev_error'))
    return error[key]

  def get_epoch_error_key_value(self, epoch):
    """
    :param int epoch:
    :return: key, error
    :rtype: (str, float)
    """
    error = self.get_epoch_error_dict(epoch)
    if not error:
      return None, None
    key = self.get_error_key(epoch)
    assert key
    assert key in error, (
      "%r not in %r. fix %r in config. set it to %r or so." %
      (key, error, 'learning_rate_control_error_measure', 'dev_error'))
    return key, error[key]

  def get_last_best_epoch(self, last_epoch, first_epoch=1, only_last_epochs=None,
                          filter_score=float("inf"),
                          only_last_n=-1,
                          min_score_dist=0.0):
    """
    :param int first_epoch: will check all epochs >= first_epoch
    :param int last_epoch: inclusive. will check all epochs <= last_epoch
    :param int|None only_last_epochs: if set, will only check the last N epochs, inclusive
    :param float filter_score: all epochs which values over this score are not considered
    :param int only_last_n: if set (>=1), *from the resulting list*, we consider only the last only_last_n
    :param float min_score_dist: filter out epochs where the diff to the most recent is not big enough
    :return: the last best epoch. to get the details then, you might want to use getEpochErrorDict.
    :rtype: int|None
    """
    if only_last_epochs:
      first_epoch = max(first_epoch, last_epoch - only_last_epochs + 1)
    if first_epoch > last_epoch:
      return None
    values = [(self.get_epoch_error_key_value(ep), ep) for ep in range(first_epoch, last_epoch + 1)]
    # Note that the order of the checks here is a bit arbitrary but I had some thoughts on it.
    # Changing the order will also slightly change the behavior, so be sure it make sense.
    values = [((key, v), ep) for ((key, v), ep) in values if v is not None]
    if not values:
      return None
    last_key, latest_score = values[-1][0]
    values = [(v, ep) for ((key, v), ep) in values if key == last_key]  # only same key
    values = [(v, ep) for (v, ep) in values if v <= filter_score]
    if not values:
      return None
    if only_last_n >= 1:
      values = values[-only_last_n:]
    values = [(v, ep) for (v, ep) in values if v + min_score_dist < latest_score]
    if not values:
      return None
    return min(values)[1]

  def save(self):
    """
    Save the current epoch data to file (self.filename).
    """
    if not self.filename:
      return
    # First write to a temp-file, to be sure that the write happens without errors.
    # Otherwise, it could happen that we delete the old existing file, then
    # some error happens (e.g. disk quota), and we loose the newbob data.
    # Loosing that data is very bad because it basically means that we have to redo all the training.
    tmp_filename = self.filename + ".new_tmp"
    f = open(tmp_filename, "w")
    f.write(better_repr(self.epoch_data))
    f.write("\n")
    f.close()
    os.rename(tmp_filename, self.filename)

  def load(self):
    """
    Loads the saved epoch data from file (self.filename).
    """
    s = open(self.filename).read()
    self.epoch_data = eval(s, {"nan": float("nan"), "inf": float("inf")}, ObjAsDict(self))


class ConstantLearningRate(LearningRateControl):
  """
  Just a constant learning rate.
  """

  need_error_info = False

  def calc_learning_rate_for_epoch(self, epoch):
    """
    Dummy constant learning rate. Returns initial learning rate.
    :type epoch: int
    :returns learning rate
    :rtype: float
    """
    while True:
      last_epoch = self.get_last_epoch(epoch)
      if last_epoch is None:
        return self.default_learning_rate
      learning_rate = self.epoch_data[last_epoch].learning_rate
      if learning_rate is None:
        epoch = last_epoch
        continue
      return learning_rate


class NewbobRelative(LearningRateControl):
  """
  If relative diff between old and new error is over some threshold, decay learning rate.
  """

  @classmethod
  def load_initial_kwargs_from_config(cls, config):
    """
    :type config: returnn.config.Config
    :rtype: dict[str]
    """
    kwargs = super(NewbobRelative, cls).load_initial_kwargs_from_config(config)
    kwargs.update({
      "relative_error_threshold": config.float('newbob_relative_error_threshold', -0.01),
    })
    return kwargs

  def __init__(self, relative_error_threshold, **kwargs):
    """
    :type relative_error_threshold: float
    """
    super(NewbobRelative, self).__init__(**kwargs)
    self.relative_error_threshold = relative_error_threshold

  def calc_learning_rate_for_epoch(self, epoch):
    """
    Newbob+ on train data.
    :type epoch: int
    :returns learning rate
    :rtype: float
    """
    last_epoch = self.get_last_epoch(epoch)
    if last_epoch is None:
      return self.default_learning_rate
    learning_rate = self.epoch_data[last_epoch].learning_rate
    if learning_rate is None:
      return self.default_learning_rate
    last2_epoch = self.get_last_epoch(last_epoch)
    if last2_epoch is None:
      return learning_rate
    relative_error = self.calc_relative_error(last2_epoch, last_epoch)
    if relative_error is None:
      return learning_rate
    learning_rate = self.calc_learning_rate_decay_or_grow(
      learning_rate, decay=relative_error > self.relative_error_threshold)
    return learning_rate


class NewbobAbs(LearningRateControl):
  """
  If absolute diff between old and new error is over some threshold, decay learning rate.
  """

  @classmethod
  def load_initial_kwargs_from_config(cls, config):
    """
    :type config: returnn.config.Config
    :rtype: dict[str]
    """
    kwargs = super(NewbobAbs, cls).load_initial_kwargs_from_config(config)
    kwargs.update({
      "error_threshold": config.float('newbob_error_threshold', -0.01),
    })
    return kwargs

  def __init__(self, error_threshold, **kwargs):
    """
    :type error_threshold: float
    """
    super(NewbobAbs, self).__init__(**kwargs)
    self.error_threshold = error_threshold

  def calc_learning_rate_for_epoch(self, epoch):
    """
    Newbob+ on train data.

    :type epoch: int
    :returns learning rate
    :rtype: float
    """
    last_epoch = self.get_last_epoch(epoch)
    if last_epoch is None:
      return self.default_learning_rate
    learning_rate = self.epoch_data[last_epoch].learning_rate
    if learning_rate is None:
      return self.default_learning_rate
    last2_epoch = self.get_last_epoch(last_epoch)
    if last2_epoch is None:
      return learning_rate
    old_key, old_error = self.get_epoch_error_key_value(last2_epoch)
    new_key, new_error = self.get_epoch_error_key_value(last_epoch)
    if old_error is None or new_error is None:
      return learning_rate
    if old_key != new_key:
      return learning_rate
    error_diff = new_error - old_error
    learning_rate = self.calc_learning_rate_decay_or_grow(learning_rate, decay=error_diff > self.error_threshold)
    return learning_rate


class NewbobMultiEpoch(LearningRateControl):
  """
  Like :class:`NewbobRelative`, but looks at the average relative error over multiple epochs.
  This is useful together with ``partition_epoch`` from :class:`Dataset`.
  """

  @classmethod
  def load_initial_kwargs_from_config(cls, config):
    """
    :type config: returnn.config.Config
    :rtype: dict[str]
    """
    kwargs = super(NewbobMultiEpoch, cls).load_initial_kwargs_from_config(config)
    kwargs.update({
      "num_epochs": config.int("newbob_multi_num_epochs", 5),
      "update_interval": config.int("newbob_multi_update_interval", config.int("newbob_multi_num_epochs", 5)),
      "relative_error_threshold": config.float('newbob_relative_error_threshold', -0.01),
      "relative_error_grow_threshold": config.float(
        'newbob_relative_error_grow_threshold', config.float('newbob_relative_error_threshold', -0.01)),
    })
    return kwargs

  def __init__(self, num_epochs, update_interval, relative_error_threshold, relative_error_grow_threshold, **kwargs):
    """
    :param int num_epochs:
    :param int update_interval:
    :param float relative_error_threshold:
    :param float relative_error_grow_threshold:
    """
    super(NewbobMultiEpoch, self).__init__(**kwargs)
    self.num_epochs = num_epochs
    assert self.num_epochs >= 1
    self.update_interval = update_interval
    assert self.update_interval >= 1
    self.relative_error_threshold = relative_error_threshold
    self.relative_error_grow_threshold = relative_error_grow_threshold

  def _calc_mean_relative_error(self, epochs):
    """
    :param list[int] epochs:
    :return: mean of relative errors
    :rtype: float|None
    """
    assert len(epochs) >= 2
    errors = [self.calc_relative_error(epochs[i], epochs[i + 1]) for i in range(len(epochs) - 1)]
    if any([e is None for e in errors]):
      return None
    return float(numpy.mean(errors))

  def _calc_recent_mean_relative_error(self, epoch):
    """
    :param int epoch:
    :return: recent mean of relative errors
    :rtype: float|None
    """
    # Take one more than numEpochs because we are looking at the diffs.
    last_epochs = self._last_epochs_for_epoch(epoch, num_epochs=self.num_epochs + 1)
    if not last_epochs:
      return None
    # We could also use the self.numEpochs limit here. But maybe this is better.
    if len(last_epochs) <= 1:
      return None
    return self._calc_mean_relative_error(last_epochs)

  def calc_learning_rate_for_epoch(self, epoch):
    """
    Newbob+ on train data.
    :type epoch: int
    :returns learning rate
    :rtype: float
    """
    learning_rate = self.get_most_recent_learning_rate(epoch)
    # We start counting epochs at 1.
    if self.update_interval > 1 and epoch % self.update_interval != 1:
      return learning_rate
    mean_relative_error = self._calc_recent_mean_relative_error(epoch)
    if mean_relative_error is None:
      return learning_rate
    learning_rate = self.calc_learning_rate_decay_or_grow(
      learning_rate,
      decay=mean_relative_error > self.relative_error_threshold,
      grow=mean_relative_error < self.relative_error_grow_threshold)
    return learning_rate


def learning_rate_control_type(type_name):
  """
  :param str type_name:
  :rtype: type[LearningRateControl]|LearningRateControl
  """
  if type_name == "constant":
    return ConstantLearningRate
  elif type_name in ("newbob", "newbob_rel", "newbob_relative"):  # Old setups expect the relative version.
    return NewbobRelative
  elif type_name == "newbob_abs":
    return NewbobAbs
  elif type_name == "newbob_multi_epoch":
    return NewbobMultiEpoch
  else:
    assert False, "unknown learning-rate-control type %s" % type_name


def load_learning_rate_control_from_config(config):
  """
  :type config: returnn.config.Config
  :rtype: LearningRateControl
  """
  control_type = config.value("learning_rate_control", "constant")
  cls = learning_rate_control_type(control_type)
  return cls.load_initial_from_config(config)


def demo():
  """
  Demo run. Given some learning rate file (with scores / existing lrs), will calculate how lrs would have been set,
  given some config.
  """
  from returnn.util import better_exchook
  better_exchook.install()
  import returnn.__main__ as rnn
  import sys
  if len(sys.argv) <= 1:
    print("usage: python %s [config] [other options] [++check_learning_rates 1]" % __file__)
    print(
      ("example usage: "
       "python %s ++learning_rate_control newbob ++learning_rate_file newbob.data ++learning_rate 0.001") % __file__)
  rnn.init_config(command_line_options=sys.argv[1:])
  # noinspection PyProtectedMember
  rnn.config._hack_value_reading_debug()
  rnn.config.update({"log": []})
  rnn.init_log()
  rnn.init_backend_engine()
  check_lr = rnn.config.bool("check_learning_rates", False)
  from returnn.pretrain import pretrain_from_config
  pretrain = pretrain_from_config(rnn.config)
  first_non_pretrain_epoch = 1
  pretrain_learning_rate = None
  if pretrain:
    first_non_pretrain_epoch = pretrain.get_train_num_epochs() + 1
  log.initialize(verbosity=[5])
  control = load_learning_rate_control_from_config(rnn.config)
  print("LearningRateControl: %r" % control)
  if not control.epoch_data:
    print("No epoch data so far.")
    return
  first_epoch = min(control.epoch_data.keys())
  if first_epoch != 1:
    print("Strange, first epoch from epoch data is %i." % first_epoch)
  print("Error key: %s from %r" % (control.get_error_key(epoch=first_epoch), control.epoch_data[first_epoch].error))
  if pretrain:
    pretrain_learning_rate = rnn.config.float('pretrain_learning_rate', control.default_learning_rate)
  max_epoch = max(control.epoch_data.keys())
  for epoch in range(1, max_epoch + 2):  # all epochs [1..max_epoch+1]
    old_learning_rate = None
    if epoch in control.epoch_data:
      old_learning_rate = control.epoch_data[epoch].learning_rate
    if epoch < first_non_pretrain_epoch:
      learning_rate = pretrain_learning_rate
      s = "Pretrain epoch %i, fixed learning rate: %s (was: %s)" % (epoch, learning_rate, old_learning_rate)
    elif 1 < first_non_pretrain_epoch == epoch:
      learning_rate = control.default_learning_rate
      s = "First epoch after pretrain, epoch %i, fixed learning rate: %s (was %s)" % (
        epoch, learning_rate, old_learning_rate)
    else:
      learning_rate = control.calc_new_learning_rate_for_epoch(epoch)
      s = "Calculated learning rate for epoch %i: %s (was: %s)" % (epoch, learning_rate, old_learning_rate)
    if learning_rate < control.min_learning_rate:
      learning_rate = control.min_learning_rate
      s += ", clipped to %s" % learning_rate
    s += ", previous relative error: %s" % control.calc_relative_error(epoch - 2, epoch - 1)
    if hasattr(control, "_calc_recent_mean_relative_error"):
      # noinspection PyProtectedMember
      s += ", previous mean relative error: %s" % control._calc_recent_mean_relative_error(epoch)
    print(s)
    if check_lr and old_learning_rate is not None:
      if old_learning_rate != learning_rate:
        print("Learning rate is different in epoch %i!" % epoch)
        sys.exit(1)
    # Overwrite new learning rate so that the calculation for further learning rates stays consistent.
    if epoch in control.epoch_data:
      control.epoch_data[epoch].learning_rate = learning_rate
    else:
      control.epoch_data[epoch] = control.EpochData(learningRate=learning_rate)
  print("Finished, last stored epoch was %i." % max_epoch)


if __name__ == "__main__":
  demo()



"""
This provides ops as wrappers around OpenFst.
"""

import os
import platform
import tensorflow as tf

returnn_dir = os.path.dirname(os.path.abspath(__file__))
openfst_dir = returnn_dir + "/extern/openfst"


def get_fst(filename):
  """
  :param str filename: to OpenFst file
  :return: TF resource handle representing the FST
  :rtype: tf.Tensor
  """
  return get_tf_mod().open_fst_load(filename=filename)


def fst_transition(fst_handle, states, inputs):
  """
  :param tf.Tensor fst_handle: via :func:`get_fst`
  :param tf.Tensor states: [batch], int32
  :param tf.Tensor inputs: [batch], int32
  :return: (next_states, output_labels, weights). next_states can be -1 if invalid. all are shape [batch].
  :rtype: (tf.Tensor,tf.Tensor,tf.Tensor)
  """
  return get_tf_mod().open_fst_transition(handle=fst_handle, states=states, inputs=inputs)


# https://www.tensorflow.org/guide/extend/op
# Also see TFUtil.TFArrayContainer for TF resources.
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/tensor.h
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/tensor_shape.h
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/tensor_types.h
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/lib/strings/str_util.h
# https://github.com/kaldi-asr/kaldi/blob/master/src/tfrnnlm/tensorflow-rnnlm.h
_src_code = """
#include <exception>
#include <limits>
#include <fst/fstlib.h>

// Defined by OpenFst and also by TensorFlow.
// Also: https://github.com/kaldi-asr/kaldi/blob/master/src/tfrnnlm/tensorflow-rnnlm.h
#undef LOG
#undef VLOG
#undef CHECK
#undef CHECK_EQ
#undef CHECK_LT
#undef CHECK_GT
#undef CHECK_LE
#undef CHECK_GE
#undef CHECK_NE
#undef DCHECK
#undef DCHECK_EQ
#undef DCHECK_LT
#undef DCHECK_GT
#undef DCHECK_LE
#undef DCHECK_GE
#undef DCHECK_NE

#include "tensorflow/core/framework/op.h"
#include "tensorflow/core/framework/op_kernel.h"
#include "tensorflow/core/framework/shape_inference.h"
#include "tensorflow/core/framework/resource_mgr.h"
#include "tensorflow/core/framework/resource_op_kernel.h"
#include "tensorflow/core/framework/tensor.h"
#include "tensorflow/core/platform/macros.h"
#include "tensorflow/core/platform/mutex.h"
#include "tensorflow/core/platform/types.h"
#include "tensorflow/core/public/version.h"

using namespace tensorflow;


REGISTER_OP("OpenFstLoad")
.Attr("filename: string")
.Attr("container: string = ''")
.Attr("shared_name: string = ''")
.Output("handle: resource")
.SetIsStateful()
.SetShapeFn(shape_inference::ScalarShape)
.Doc("OpenFstLoad: loads FST, creates TF resource, persistent across runs in the session");


REGISTER_OP("OpenFstTransition")
.Input("handle: resource")
.Input("states: int32")
.Input("inputs: int32")
.Output("next_states: int32")
.Output("output_labels: int32")
.Output("weights: float32")
.SetShapeFn([](::tensorflow::shape_inference::InferenceContext* c) {
  c->set_output(0, c->input(1));
  c->set_output(1, c->input(1));
  c->set_output(2, c->input(1));
  return Status::OK();
})
.Doc("OpenFstTransition: performs a transition");


struct OpenFstInstance : public ResourceBase {
  typedef fst::StdArc Arc;  // FSTs are usually saved using this type, and we have to use the same
  typedef fst::VectorFst<Arc> Fst;

  explicit OpenFstInstance(const string& filename)
      : filename_(filename), fst_(Fst::Read(filename)) {
    if(!fst_)
      throw std::runtime_error("failed to load FST; see stdout for errors");
  }

  virtual ~OpenFstInstance() {
    delete fst_;
  }

  string DebugString()
#if (TF_MAJOR_VERSION == 1 && TF_MINOR_VERSION >= 14) || (TF_MAJOR_VERSION > 1)
const
#endif
  override {
    return strings::StrCat("OpenFstInstance[", filename_, "]");
  }

  // This assumes a deterministic FST, i.e. it either has a single matching arc, or none.
  bool transition(int curState, int inputLabel, int* nextState, int* outputLabel, float* weight) {
    fst::Matcher<Fst> matcher(fst_, fst::MATCH_INPUT);
    if(curState >= 0)
      matcher.SetState(curState);
    if(curState >= 0 && matcher.Find(inputLabel)) {
      const Arc& arc = matcher.Value();
      *nextState = arc.nextstate;
      *outputLabel = arc.olabel;
      *weight = arc.weight.Value();
      return true;
    }
    else {
      *nextState = -1;
      *outputLabel = -1;
      *weight = -std::numeric_limits<float>::infinity();
      return false;
    }
  }

  const string filename_;
  mutex mu_;
  Fst* fst_;
};


// https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/resource_op_kernel.h
// TFUtil.TFArrayContainer
class OpenFstLoadOp : public ResourceOpKernel<OpenFstInstance> {
 public:
  explicit OpenFstLoadOp(OpKernelConstruction* context)
      : ResourceOpKernel(context) {
    OP_REQUIRES_OK(context, context->GetAttr("filename", &filename_));
  }

 private:
  virtual bool IsCancellable() const { return false; }
  virtual void Cancel() {}

  Status CreateResource(OpenFstInstance** ret) override {
    try {
      *ret = new OpenFstInstance(filename_);
    } catch (std::exception& exc) {
      return errors::Internal("Could not load OpenFst ", filename_, ", exception: ", exc.what());
    }
    if(*ret == nullptr)
      return errors::ResourceExhausted("Failed to allocate");
    return Status::OK();
  }

  Status VerifyResource(OpenFstInstance* fst) override {
    if(fst->filename_ != filename_)
      return errors::InvalidArgument("Filename mismatch: expected ", filename_,
                                     " but got ", fst->filename_, ".");
    return Status::OK();
  }

  string filename_;
};

REGISTER_KERNEL_BUILDER(Name("OpenFstLoad").Device(DEVICE_CPU), OpenFstLoadOp);


class OpenFstTransitionOp : public OpKernel {
 public:
  using OpKernel::OpKernel;

  void Compute(OpKernelContext* context) override {
    OpenFstInstance* fst;
    OP_REQUIRES_OK(context, GetResourceFromContext(context, "handle", &fst));
    core::ScopedUnref unref(fst);

    const Tensor& states_tensor = context->input(1);
    auto states_flat = states_tensor.flat<int32>();

    const Tensor& inputs_tensor = context->input(2);
    auto inputs_flat = inputs_tensor.flat<int32>();

    OP_REQUIRES(
      context,
      TensorShapeUtils::IsVector(states_tensor.shape()) &&
      TensorShapeUtils::IsVector(inputs_tensor.shape()) &&
      states_flat.size() == inputs_flat.size(),
      errors::InvalidArgument(
        "Shape mismatch. states ", states_tensor.shape().DebugString(),
        " vs inputs ", inputs_tensor.shape().DebugString()));

    Tensor* output_next_states_tensor = NULL;
    OP_REQUIRES_OK(context, context->allocate_output(0, states_tensor.shape(), &output_next_states_tensor));
    auto output_next_states_flat = output_next_states_tensor->flat<int32>();
    Tensor* output_output_labels_tensor = NULL;
    OP_REQUIRES_OK(context, context->allocate_output(1, states_tensor.shape(), &output_output_labels_tensor));
    auto output_output_labels_flat = output_output_labels_tensor->flat<int32>();
    Tensor* output_weights_tensor = NULL;
    OP_REQUIRES_OK(context, context->allocate_output(2, states_tensor.shape(), &output_weights_tensor));
    auto output_weights_flat = output_weights_tensor->flat<float>();

    for(int i = 0; i < inputs_flat.size(); ++i) {
      fst->transition(
        states_flat(i),
        inputs_flat(i),
        &output_next_states_flat(i),
        &output_output_labels_flat(i),
        &output_weights_flat(i));
    }
  }
};

REGISTER_KERNEL_BUILDER(Name("OpenFstTransition").Device(DEVICE_CPU), OpenFstTransitionOp);

"""


def openfst_checked_out():
  """
  :return: whether the Git submodule is checked out
  :rtype: bool
  """
  return os.path.exists("%s/src/include/fst/fst.h" % openfst_dir)


_tf_mod = None


def get_tf_mod(verbose=False):
  """
  :param bool verbose:
  :return: module
  """
  global _tf_mod
  if _tf_mod:
    return _tf_mod
  from glob import glob
  from returnn.tf.util.basic import OpCodeCompiler

  # Collect files.
  assert openfst_checked_out(), "submodule in %r not checked out?" % openfst_dir
  files = glob('%s/src/lib/*.cc' % openfst_dir)
  assert files, "submodule in %r not checked out?" % openfst_dir
  files = sorted(files)  # make somewhat deterministic
  libs = []
  if platform.system() != 'Darwin':
    libs.append('rt')

  # Put code all together in one big blob.
  src_code = ""
  for fn in files:
    f_code = open(fn).read()
    f_code = ''.join([x for x in f_code if ord(x) < 128])  # enforce ASCII
    src_code += "\n// ------------ %s : BEGIN { ------------\n" % os.path.basename(fn)
    # https://gcc.gnu.org/onlinedocs/cpp/Line-Control.html#Line-Control
    src_code += "#line 1 \"%s\"\n" % os.path.basename(fn)
    src_code += f_code
    src_code += "\n// ------------ %s : END } --------------\n\n" % os.path.basename(fn)
  src_code += "\n\n// ------------ our code now: ------------\n\n"
  src_code += "#line 1 \"returnn/tf/util/open_fst.py:_src_code\"\n"
  src_code += _src_code

  compiler = OpCodeCompiler(
    base_name="OpenFst", code_version=1, code=src_code,
    include_paths=("%s/src/include" % openfst_dir,),
    c_macro_defines={
      "NDEBUG": 1,  # https://github.com/tensorflow/tensorflow/issues/17316
    },
    ld_flags=["-l%s" % lib for lib in libs],
    is_cpp=True, use_cuda_if_available=False,
    verbose=verbose)
  tf_mod = compiler.load_tf_module()
  assert hasattr(tf_mod, "open_fst_transition"), "content of mod: %r" % (dir(tf_mod),)
  _tf_mod = tf_mod
  return tf_mod


def _demo():
  def _make_int_list(s):
    """
    :param str s:
    :rtype: list[int]
    """
    return [int(s_) for s_ in s.split(",")]

  from returnn.util import better_exchook
  better_exchook.install()
  from argparse import ArgumentParser
  arg_parser = ArgumentParser()
  arg_parser.add_argument("--states", type=_make_int_list, default=[0])
  arg_parser.add_argument("--inputs", type=_make_int_list, default=[0])
  # Other example FST files can be found online, e.g.: https://github.com/placebokkk/gofst/tree/master/ex01
  # Or: https://github.com/zh794390558/deeplearning/tree/master/kaldi/fst/data/test
  arg_parser.add_argument("--fst", default=returnn_dir + "/tests/lexicon_opt.fst")
  args = arg_parser.parse_args()
  # Try to compile now.
  get_tf_mod(verbose=True)
  # Some demo.
  assert os.path.exists(args.fst)
  fst_tf = get_fst(filename=args.fst)
  states_tf = tf.compat.v1.placeholder(tf.int32, [None])
  inputs_tf = tf.compat.v1.placeholder(tf.int32, [None])
  output_tf = fst_transition(fst_handle=fst_tf, states=states_tf, inputs=inputs_tf)
  with tf.compat.v1.Session() as session:
    out_next_states, out_labels, out_scores = session.run(
      output_tf, feed_dict={
        states_tf: args.states, inputs_tf: args.inputs})
    print("states:", args.states)
    print("inputs:", args.inputs)
    print("output next states:", out_next_states)
    print("output labels:", out_labels)
    print("output scores:", out_scores)


if __name__ == "__main__":
  _demo()




"""
Provide basic Byte-Pair-Encoding (BPE) utilities.
"""

import re
import typing
import numpy


BpeMergeSymbol = "@@"


class StandardBytePairEncoder:
  """
  Code is partly taken from subword-nmt/apply_bpe.py.
  Author: Rico Sennrich, code under MIT license.

  Use operations learned with learn_bpe.py to encode a new text.
  The text will not be smaller, but use only a fixed vocabulary, with rare words
  encoded as variable-length sequences of subword units.

  Reference:
  Rico Sennrich, Barry Haddow and Alexandra Birch (2016). Neural Machine Translation of Rare Words with Subword Units.
  Proceedings of the 54th Annual Meeting of the Association for Computational Linguistics (ACL 2016). Berlin, Germany.

  """

  def __init__(self, bpe_codes_file, labels=None):
    """
    :param str bpe_codes_file: codes file
    :param list[str]|None labels: vocab
    """
    self.labels = labels
    self._load(bpe_codes_file)
    self._bpe_encode_cache = {}
    self._bpe_separator = BpeMergeSymbol

  _file_cache = {}  # filename -> bpe_file_version, bpe_codes, bpe_codes_reverse

  def _load(self, bpe_codes_file):
    """
    Load BPE codes from file

    :param str bpe_codes_file: codes file
    """
    if bpe_codes_file in self._file_cache:
      self._bpe_file_version, self._bpe_codes, self._bpe_codes_reverse = self._file_cache[bpe_codes_file]
      return
    # check version information
    bpe_file_first_line = open(bpe_codes_file, "r").readline()
    if bpe_file_first_line.startswith('#version:'):
      self._bpe_file_version = tuple(
        [int(x) for x in re.sub(r'(\.0+)*$', '', bpe_file_first_line.split()[-1]).split(".")])
    else:
      self._bpe_file_version = (0, 1)
    self._bpe_codes = [tuple(item.split()) for item in open(bpe_codes_file, "rb").read().decode("utf8").splitlines()]
    # some hacking to deal with duplicates (only consider first instance)
    self._bpe_codes = dict([(code, i) for (i, code) in reversed(list(enumerate(self._bpe_codes)))])
    self._bpe_codes_reverse = dict([(pair[0] + pair[1], pair) for pair, i in self._bpe_codes.items()])
    self._file_cache[bpe_codes_file] = (self._bpe_file_version, self._bpe_codes, self._bpe_codes_reverse)

  @staticmethod
  def _get_pairs(word):
    """
    :param tuple[str] word: represented as tuple of symbols (symbols being variable-length strings)
    :return: set of symbol pairs in a word
    :rtype: set[(str,str)]
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
      pairs.add((prev_char, char))
      prev_char = char
    return pairs

  def _encode_word(self, orig):
    """
    Encode word based on list of BPE merge operations, which are applied consecutively.
    :param str orig:
    :rtype: tuple[str]
    """

    if orig in self._bpe_encode_cache:
      return self._bpe_encode_cache[orig]

    if self._bpe_file_version == (0, 1):
      word = tuple(orig) + ('</w>',)
    elif self._bpe_file_version == (0, 2):  # more consistent handling of word-final segments
      word = tuple(orig[:-1]) + (orig[-1] + '</w>',)
    else:
      raise NotImplementedError

    pairs = self._get_pairs(word)
    if not pairs:
      return orig

    while True:
      bigram = min(pairs, key=lambda pair: self._bpe_codes.get(pair, float('inf')))
      if bigram not in self._bpe_codes:
        break
      first, second = bigram
      new_word = []
      i = 0
      while i < len(word):
        try:
          j = word.index(first, i)
          new_word.extend(word[i:j])
          i = j
        except ValueError:
          new_word.extend(word[i:])
          break

        if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
          new_word.append(first + second)
          i += 2
        else:
          new_word.append(word[i])
          i += 1
      new_word = tuple(new_word)
      word = new_word
      if len(word) == 1:
        break
      else:
        pairs = self._get_pairs(word)

    # don't print end-of-word symbols
    if word[-1] == '</w>':
      word = word[:-1]
    elif word[-1].endswith('</w>'):
      word = word[:-1] + (word[-1].replace('</w>', ''),)

    if self.labels:
      word = self._check_vocab_and_split(word, self._bpe_codes_reverse, self.labels, self._bpe_separator)

    self._bpe_encode_cache[orig] = word
    return word

  def _check_vocab_and_split(self, orig, bpe_codes, vocab, separator):
    """
    Check for each segment in word if it is in-vocabulary,
    and segment OOV segments into smaller units by reversing the BPE merge operations
    """

    out = []

    for segment in orig[:-1]:
      if segment + separator in vocab:
        out.append(segment)
      else:
        # sys.stderr.write('OOV: {0}\n'.format(segment))
        for item in self._recursive_split(segment, bpe_codes, vocab, separator, False):
          out.append(item)

    segment = orig[-1]
    if segment in vocab:
      out.append(segment)
    else:
      # sys.stderr.write('OOV: {0}\n'.format(segment))
      for item in self._recursive_split(segment, bpe_codes, vocab, separator, True):
        out.append(item)

    return out

  def _recursive_split(self, segment, bpe_codes, vocab, separator, final=False):
    """Recursively split segment into smaller units (by reversing BPE merges)
    until all units are either in-vocabulary, or cannot be split further."""

    # noinspection PyBroadException
    try:
      if final:
        left, right = bpe_codes[segment + '</w>']
        right = right[:-4]
      else:
        left, right = bpe_codes[segment]
    except Exception:  # TODO fix
      # sys.stderr.write('cannot split {0} further.\n'.format(segment))
      yield segment
      return

    if left + separator in vocab:
      yield left
    else:
      for item in self._recursive_split(left, bpe_codes, vocab, separator, False):
        yield item

    if (final and right in vocab) or (not final and right + separator in vocab):
      yield right
    else:
      for item in self._recursive_split(right, bpe_codes, vocab, separator, final):
        yield item

  def segment_sentence(self, sentence):
    """
    Segment single sentence (whitespace-tokenized string) with BPE encoding.

    :param str sentence:
    :rtype: list[str]
    """

    output = []

    found_category = False
    skip_category = False

    for word in sentence.split():
      if word[0] == '$' and len(word) > 1:
        found_category = True
        output.append(word)
      elif found_category is True and word[0] == '{':
        skip_category = True
        output.append(word)
      elif skip_category is True and word[0] != '}':
        output.append(word)
      else:
        found_category = False
        skip_category = False
        new_word = self._encode_word(word)

        for item in new_word[:-1]:
          output.append(item + self._bpe_separator)
        output.append(new_word[-1])

    return output


class PrefixTree:
  """
  Prefix tree / trie.
  This class represents both a single node and the tree.
  """

  def __init__(self, prefix="", root=None):
    """
    :param str prefix:
    :param PrefixTree|None root:
    """
    self.prefix = prefix
    self.arcs = {}  # type: typing.Dict[str,PrefixTree]
    self.finished = False
    self.bpe_finished = False
    self.is_root = not root
    self.root = root

  def add(self, postfix, root=None):
    """
    :param str postfix:
    :param None|PrefixTree root:
    :rtype: PrefixTree
    """
    if not root:
      if self.is_root:
        root = self
      else:
        assert self.root
        root = self.root
    if postfix == BpeMergeSymbol:
      arc = postfix
      postfix_ = ""
    else:
      arc = postfix[:1]
      postfix_ = postfix[1:]
    if arc in self.arcs:
      child = self.arcs[arc]
    else:
      child = PrefixTree(root=root, prefix=self.prefix + arc)
      self.arcs[arc] = child
    if arc == BpeMergeSymbol and not postfix_:
      self.bpe_finished = True
    if postfix_:
      return child.add(postfix_, root=root)
    else:
      child.finished = True
      return child


class Hyp:
  """
  Represents a hypothesis in the search.
  """

  def __init__(self, bpe_sym_history, cur_node):
    """
    :param list[str] bpe_sym_history:
    :param PrefixTree cur_node:
    """
    self.bpe_sym_history = bpe_sym_history
    self.cur_node = cur_node


class CharSyncSearch:
  """
  Covers the search hyps and the search itself.
  """

  def __init__(self, bpe, word, word_pos=0):
    """
    :param PrefixTree bpe:
    :param str word:
    :param int word_pos:
    """
    self.bpe = bpe
    self.word = word
    self.word_pos = word_pos
    self.hyps = [Hyp(bpe_sym_history=[], cur_node=bpe)]  # type: typing.List[Hyp]
    self.final_bpe_seqs = None  # type: typing.Optional[typing.List[typing.List[str]]]

  def _get_finished(self):
    assert self.word_pos == len(self.word)
    finals = []  # type: typing.List[typing.List[str]]
    for hyp in self.hyps:
      if hyp.cur_node.finished:
        finals.append(hyp.bpe_sym_history + [hyp.cur_node.prefix])
    self.final_bpe_seqs = finals

  def _expand(self):
    assert self.word_pos < len(self.word)
    char = self.word[self.word_pos]
    new_hyps = []  # type: typing.List[Hyp]
    for hyp in self.hyps:
      if hyp.cur_node.bpe_finished:
        next_node = self.bpe.arcs.get(char)
        if next_node:
          new_hyps.append(
            Hyp(
              bpe_sym_history=hyp.bpe_sym_history + [hyp.cur_node.prefix + BpeMergeSymbol],
              cur_node=next_node))
      next_node = hyp.cur_node.arcs.get(char)
      if next_node:
        new_hyps.append(Hyp(bpe_sym_history=hyp.bpe_sym_history, cur_node=next_node))
    self.hyps = new_hyps

  def search(self):
    """
    :return: collection of possible BPE symbol seqs
    :rtype: list[list[str]]
    """
    while self.word_pos < len(self.word):
      self._expand()
      self.word_pos += 1
    self._get_finished()
    return self.final_bpe_seqs


class HypInPos:
  """
  Represents a hypothesis in the search.
  """

  def __init__(self, bpe_sym_history, cur_node, pos):
    """
    :param list[str] bpe_sym_history:
    :param PrefixTree cur_node:
    :param int pos:
    """
    self.bpe_sym_history = bpe_sym_history
    self.cur_node = cur_node
    self.pos = pos


class DepthFirstSearch:
  """
  Depth-first search.
  """

  def __init__(self, bpe, word, sampler=None):
    """
    :param PrefixTree bpe:
    :param str word:
    :param (()->bool)|None sampler:
    """
    self.bpe = bpe
    self.word = word
    self.sampler = sampler
    self.hyps = []  # type: typing.List[HypInPos]
    self.final_bpe_seq = None  # type: typing.Optional[typing.List[str]]
    self._add_hyp(HypInPos(bpe_sym_history=[], cur_node=bpe, pos=0))

  def _add_hyp(self, hyp):
    """
    :param HypInPos hyp:
    """
    if hyp.pos >= len(self.word):
      if hyp.cur_node.finished:
        self.final_bpe_seq = hyp.bpe_sym_history + [hyp.cur_node.prefix]
    else:
      self.hyps.append(hyp)

  def _expand(self):
    hyp = self.hyps.pop(-1)

    # Now check for possible further hyps.
    char = self.word[hyp.pos]
    new_hyps = []  # type: typing.List[HypInPos]
    if hyp.cur_node.bpe_finished:
      next_node = self.bpe.arcs.get(char)
      if next_node:
        new_hyps.append(
          HypInPos(
            bpe_sym_history=hyp.bpe_sym_history + [hyp.cur_node.prefix + BpeMergeSymbol],
            cur_node=next_node, pos=hyp.pos + 1))
    next_node = hyp.cur_node.arcs.get(char)
    if next_node:
      new_hyps.append(HypInPos(bpe_sym_history=hyp.bpe_sym_history, cur_node=next_node, pos=hyp.pos + 1))

    # Note that the order we check them will make this a depth-first or breadth-first search.
    if self.sampler and self.sampler():
      new_hyps = list(reversed(new_hyps))
    for hyp in new_hyps:
      self._add_hyp(hyp)

  def search(self):
    """
    :return: BPE symbol seq if one is found
    :rtype: list[str]|None
    """
    while self.hyps and self.final_bpe_seq is None:
      self._expand()
    return self.final_bpe_seq


class SamplingBytePairEncoder:
  """
  Will randomly sample from any possible BPE split.
  """

  def __init__(self, labels, breadth_prob, rnd, unknown_label=None):
    """
    :param list[str] labels: vocab
    :param float breadth_prob: 1.0 will lead to breadth-first search, 0.0 to depth-first search.
      other values are stochastic.
    :param numpy.random.RandomState rnd:
    :param str|None unknown_label:
    """
    self.labels = labels
    self.unknown_label = unknown_label
    if unknown_label:
      assert unknown_label in self.labels
    self.breadth_prob = breadth_prob
    self.rnd = rnd

    # build prefix tree
    bpe = PrefixTree()
    for bpe_sym in labels:
      bpe.add(bpe_sym)
    self._bpe_prefix_tree = bpe

  def _sampler(self):
    # When this returns true, it will differ from depth-first search.
    return self.rnd.random_sample() <= self.breadth_prob

  def get_bpe_split_for_word(self, word):
    """
    :param str word:
    :rtype: list[str]|None
    """
    return DepthFirstSearch(bpe=self._bpe_prefix_tree, word=word, sampler=self._sampler).search()

  def segment_sentence(self, sentence):
    """
    Segment single sentence (whitespace-tokenized string) with BPE encoding.

    :param str sentence:
    :rtype: list[str]
    """
    output = []
    for word in sentence.split():
      bpe_sym_seq = self.get_bpe_split_for_word(word)
      if bpe_sym_seq is None:
        if self.unknown_label:
          output.append(self.unknown_label)
          continue
        else:
          raise Exception("no BPE-split for word %r" % word)
      output.extend(bpe_sym_seq)
    return output


def _demo():
  import sys
  import os
  my_dir = os.path.dirname(os.path.abspath(__file__))
  root_dir = os.path.dirname(os.path.dirname(my_dir))
  assert os.path.exists("%s/returnn/__init__.py" % root_dir)
  sys.path.insert(0, root_dir)

  from returnn.util import better_exchook
  better_exchook.install()

  import argparse
  arg_parser = argparse.ArgumentParser()
  arg_parser.add_argument("--vocab", required=True)
  arg_parser.add_argument("--unk")
  arg_parser.add_argument("--input", help="text. if not given, will read from stdin")
  arg_parser.add_argument("--seed", type=int, default=0)
  arg_parser.add_argument("--all", action="store_true")
  arg_parser.add_argument(
    "--breadth-prob", type=float, default=0.0,
    help="1.0 will lead to breadth-first search, 0.0 to depth-first search. other values are stochastic.")
  args = arg_parser.parse_args()

  from returnn.datasets.util.vocabulary import Vocabulary
  vocab = Vocabulary(vocab_file=args.vocab, unknown_label=None)
  rnd = numpy.random.RandomState(args.seed)

  if args.input:
    bpe_prefix_tree = PrefixTree()
    for bpe_sym in vocab.labels:
      bpe_prefix_tree.add(bpe_sym)

    def _sampler():
      # When this returns true, it will differ from depth-first search.
      return rnd.random_sample() <= args.breadth_prob

    for word in args.input.split():
      if args.all:
        bpe_sym_seqs = CharSyncSearch(bpe=bpe_prefix_tree, word=word).search()
        print("%s: %s" % (word, bpe_sym_seqs))
      else:
        greedy = DepthFirstSearch(bpe=bpe_prefix_tree, word=word, sampler=_sampler).search()
        print("%s: %s" % (word, " ".join(greedy)))
    return

  bpe = SamplingBytePairEncoder(labels=vocab.labels, breadth_prob=args.breadth_prob, rnd=rnd, unknown_label=args.unk)
  print("Reading from stdin:")
  while True:
    try:
      line = sys.stdin.readline()
      if line == "":  # EOF
        return
    except KeyboardInterrupt:
      return
    line = line.strip()
    print(" ".join(bpe.segment_sentence(line)))


if __name__ == "__main__":
  _demo()
